/* 
  Jeenode Transmit sketch for model rocket telemetry
    - Transmits time, acceration and pressure as rapidly 
      as possible and stores values to memory if available.
  
  648.ken@gmail.com,  bugled.blogspot.com
    
  Jeenode V5 w/ Atmega328p
    Port 3:
      Gravity Plug (BMA020 3-axis digital accelerometer)
      Pressure Plug (Bosch BMP085 digital pressure detector)
      Memory Plug (1 x M24M01 128Kbyte EEPROM)
    Port 4:
      LED + resistor
*/

// includes  
#include <Ports.h> //jeelabs.net/projects/cafe/wiki/Libraries
#include <RF12.h> 
#include "PortsBMP085.h"

// hardware setup
PortI2C myBus (3);
GravityPlug sensor (myBus); 
BMP085 psensor (myBus, 0); // low power
MemoryPlug mem (myBus);
Port LED (4);

// variables
boolean DEBUG = false; // print debugging messages
boolean WRITE = false;  // write values to memory plug
int range; // gravity plug scale (2, 4, or 8g)
int accel_axis = 2; // 0 = x, 1 = y, 2 = z
int16_t inital_accel;
int32_t initial_pres;
int count;
int ledState = LOW;

typedef struct { // tranmission data (6 bytes)
        uint16_t ms;
	int16_t accel;
	int16_t pres;
} Payload;
Payload values;

typedef struct { // memory plug pointer
  word page; // 0 - 500 pages
  byte offset; // 0 - 255 bytes
} Mem_location;
Mem_location nextWrite;

struct { int16_t temp; int32_t pres; } presVals;


void setup () {
    Serial.flush();  
    Serial.begin(57600);
    sensor.begin();
    psensor.getCalibData();
    rf12_initialize(2, RF12_915MHZ, 212); // 915 Mhz, net group 212, node 2
    if (DEBUG) {Serial.print  ("RM12B config: ");}
    rf12_config(); // ID of RF12 node
    /*
        range<1:0>
        B00 = +/- 2g <-- default     (cal = 128)
        B01 = +/- 4g
        B10 = +/- 8g                 (cal = 144)
    */
    //sensor.setRange(B10);
    sensor.setSettings(144);
    range = sensor.getRange();
    LED.mode(OUTPUT);

    initialize(); // zero sensors
    
    getNextMem();
    if (nextWrite.page > 499){
      Serial.println("DATA MEMORY FULL");
      clearMem();
      if (DEBUG){delay(500);}
    }
 }


void loop(){
  
  checkSerial(); // look for incomming
  if (count == 0){ // start a BMP085 measurment
    psensor.startMeas(BMP085::TEMP);}
  else {
    psensor.startMeas(BMP085::PRES);}
    
  int accel_count = 0;
  uint32_t startTime = millis();
  int accel_total = 0;
  while(millis() - startTime < 5){ // check accel
    const int* p = sensor.getAxes(); // read x,y,z data
    accel_total = accel_total + p[accel_axis]; // y axis  
    accel_count ++;
  }
  checkRecieved();
  values.accel = inital_accel - (accel_total / accel_count);  
  if (count == 0) {
    int32_t traw = psensor.getResult(BMP085::TEMP);
  }
  else {
    int32_t praw = psensor.getResult(BMP085::PRES);
    psensor.calculate(presVals.temp, presVals.pres);
    values.pres = initial_pres - presVals.pres;  
  }
  
  uint32_t curTime = millis();
  uint32_t gtTime = round(curTime / 60000) * 60000;
  values.ms = curTime - gtTime;
  if (WRITE && DEBUG){
    Serial.print(" time = ");
    Serial.print(values.ms);  
    Serial.print(" accel = ");
    Serial.print(values.accel);
    Serial.print(" temp = ");
    Serial.print(presVals.temp);
    Serial.print(" press = ");
    Serial.println(values.pres);
  }
  checkRecieved();
  if (WRITE){
    writeMem();
    if (rf12_canSend()){ // transmit if ready
      rf12_sendStart(0, &values, sizeof values, 1); }
  }

  count ++;
  if (count > 100) { count = 0;} // get temp every 100 loops
}


void initialize(){
    // get inital readings (assumes at rest at ground level)
    const int* p = sensor.getAxes(); // read x,y,z data
    inital_accel = p[accel_axis];
    psensor.startMeas(BMP085::TEMP);
    delay(5); // min time to get reading
    int32_t traw = psensor.getResult(BMP085::TEMP);
    psensor.startMeas(BMP085::PRES);
    delay(5); // min time to get reading in low power mode
    int32_t praw = psensor.getResult(BMP085::PRES);
    psensor.calculate(presVals.temp, presVals.pres);
    initial_pres = presVals.pres;
    values.ms = 0;
    values.pres = 0;
    values.accel = 0;
    rf12_sendStart(0, &values, sizeof values, 1);

    if (true) { // print setup & data
      Serial.println("\n[Transmit.pde]");
      Serial.print("\taccel range = ");
      Serial.println(range, DEC);    
      Serial.print("\tconfig bit = ");
      Serial.println(sensor.getSettings(), DEC);
      Serial.print("\taccel axis = ");
      Serial.println(accel_axis, DEC);      
      Serial.print("\tinitial accel = ");
      Serial.println(inital_accel, DEC);
      Serial.print("\tinitial temp = ");
      Serial.println(presVals.temp, DEC);
      Serial.print("\tinitial pres = ");
      Serial.println(initial_pres, DEC);
    }
    flashLED();
    count = 0;
}


void flashLED(){ // flash LED on & off
    LED.digiWrite(1);
    delay(250);
    LED.digiWrite(0);
    delay(250);
}


void writeMem(){ // write payload values to memory
  if (mem.isPresent()) {
    mem.save(nextWrite.page, &values, nextWrite.offset, sizeof values);
    if (nextWrite.page < 500){
      if (nextWrite.offset < (255 - 2 * sizeof values)){
        nextWrite.offset = nextWrite.offset + sizeof values;
      }
      else{
        nextWrite.offset = 0;
        nextWrite.page = nextWrite.page + 1;
      }
      mem.save(0, &nextWrite, 252, sizeof nextWrite);
    }
  }
}


void getNextMem(){ // get memory position to write to next
  uint8_t buf[] = { 0, 0, 0 };
  //delay(10);
  if (mem.isPresent()) {
    mem.load(0, &buf, 252, 3); // location of "last write" (end of first page)
    nextWrite.page = buf[1] << 8 | buf[0];
    nextWrite.offset = buf[2];
    Serial.print("\tmem: page = ");
    Serial.print(nextWrite.page, DEC);
    Serial.print(" offset = ");
    Serial.println(nextWrite.offset, DEC);
  }
  else {
    Serial.println("MEMORY NOT PRESENT");
    if (DEBUG) { delay(500);}
  }
}


void clearMem(){ // sets memory pointer to zero
  if (mem.isPresent()) {
    flashLED();
    uint8_t buf[] = { 0, 0, 0 }; // start right after address
    mem.save(0, &buf, 252, 3);
    nextWrite.page = 0;
    nextWrite.offset = 0;
    Serial.println("MEMORY CLEARED");
    //getNextMem();
    initialize();
    if (DEBUG) {delay(500);}
  }
  else {
    Serial.println("MEMORY NOT PRESENT");
  }

}


void readMem(){ // read contents of memory
  if (mem.isPresent()) {
    if (DEBUG) {Serial.println("Start OF MEMORY READ");}
    flashLED();
    for(word curPage = 0; curPage <= nextWrite.page; curPage ++){
      for(int curOffset = 0; curOffset < (255 - sizeof values); curOffset = curOffset + sizeof values){
        if (curPage == nextWrite.page && curOffset == nextWrite.offset) { break;}
        mem.load(curPage, &values, curOffset, sizeof values);
        Serial.print(curPage);
        Serial.print("-");
        Serial.print(curOffset);
        Serial.print(":");
        Serial.print(values.ms);
        Serial.print(' ');
        Serial.print(values.accel);
        Serial.print(' ');
        Serial.println(values.pres);

        // check incomming, send outgoing
        checkRecieved();
        while (!rf12_canSend())
          if (rf12_recvDone()){ checkRecieved();}
        rf12_sendStart(0, &values, sizeof values, 1);
        delay(5); 
      }
    }
    flashLED();
    if (DEBUG) {Serial.println("END OF MEMORY READ");}
  }
  else {
    Serial.println("MEMORY NOT PRESENT");
  }
  if (DEBUG) {delay(500);}
}  
  
  
void checkSerial(){ // check for incomming serial data
  if (Serial.available()>0){
    int incomingByte = Serial.read();
    if (incomingByte == 99){ // "c"
      clearMem();
    }
    else if (incomingByte == 100){ // "d"
      DEBUG = !DEBUG;
      Serial.print("DEBUG=");
      Serial.println(DEBUG, DEC);
    }
    else if (incomingByte == 119){ //"w"
       WRITE = !WRITE;
       if (WRITE) {
         Serial.println("WRITING");
         ledState = HIGH;
         LED.digiWrite(ledState);
       }
       else {
         Serial.print("STOP WRITING @ ");
         Serial.print(nextWrite.page, DEC);
         Serial.print(" - ");
         Serial.println(nextWrite.offset, DEC);
         ledState = LOW;
         LED.digiWrite(ledState);
       }
    }
    else if (incomingByte == 114){ // "r"
      readMem();
    }
  }
}


void checkRecieved(){ // check incomming transmission data
    byte n = rf12_len;
    if (rf12_recvDone() && rf12_crc == 0 && rf12_len > 0){
      Serial.print("incomming -> ");
      Serial.print((int) rf12_hdr);
      for (byte i = 0; i < n; ++i) {
          Serial.print(' ');
          Serial.print((int) rf12_data[i]);
        }
      //int incommingByte = rf12_data[0];
      //while (!rf12_canSend()){
        //rf12_recvDone();}
      //rf12_sendStart(0, &incommingByte, 1);
     // delay(500);
      Serial.println(); 
      if ((int) rf12_data[0] == 114){
        Serial.println("Read Mem");
        readMem();
      }      
      else if ((int) rf12_data[0] == 119){
        Serial.println("Write mem");
        WRITE = !WRITE;
        if (WRITE) {
          Serial.println("WRITING");
          ledState = HIGH;
          LED.digiWrite(ledState);
        }
        else {
          Serial.print("STOP WRITING @ ");
          Serial.print(nextWrite.page, DEC);
          Serial.print(" - ");
          Serial.println(nextWrite.offset, DEC);
          ledState = LOW;
          LED.digiWrite(ledState);
        }
      }
      else if ((int) rf12_data[0] == 99){
        Serial.println("Clear mem");
        clearMem();
      }
    } 
  }

